import React, { MutableRefObject, useEffect } from 'react'
import { BaseEditor, createEditor, Editor, Node, Transforms } from 'slate'
import { withHistory } from 'slate-history'
import { DefaultElement, Editable, ReactEditor, Slate, withReact } from 'slate-react'
import { ICompletionResponse } from './ICompletionResponse'
import { SmartInputFetcher } from './SmartInputFetcher'

/**
 * Properties for the `SmartInput` component.
 */
export interface SmartInputProps {
  /**
   * The controlled value of the input. Declare a state variable for this in the parent component.
   */
  value: string
  /**
   * Called whenever the value of the input changes. Use this to update the state variable for the controlled `value` property.
   */
  onChange: (newValue: string) => void
  /**
   * Renders text the user has input.
   */
  renderText: (props: any) => React.ReactElement<any, any>
  /**
   * Renders suggested, tab-completable text.
   */
  renderCompletion: (props: any) => React.ReactElement<any, any>
  /**
   * Renders the placeholder text when the input value is empty.
   * If not provided, the placeholder will be rendered with a regular, low opacity text node.
   */
  renderPlaceholder?: (props: any) => React.ReactElement<any, any>
  /**
   * The ID of the model to use for completions. If not provided, a sample completion will be displayed.
   */
  modelId?: string
  /**
   * An object that will be passed to the model when fetching completions. This can be used to provide context for the completions.
   * Pass anything that you think is relevant to making the completions better or more personalized (e.g. the current user id, additional form data, etc.)
   */
  context?: any
  /**
   * A ref to the underlying contenteditable div. This can be used to focus the editor, select text, etc.
   */
  editorRef?: MutableRefObject<any>
  /**
   * If true, the input value may contain newline characters. If false, the input value will be foreced to stay a single line.
   */
  multiline?: boolean
  /**
   * Called when the input loses focus.
   */
  onBlur?: React.FocusEventHandler<HTMLDivElement>
  /**
   * Called when the input gains focus.
   */
  onFocus?: React.FocusEventHandler<HTMLDivElement>
  /**
   * Called when a key is pressed while the input is focused.
   */
  onKeyDown?: React.KeyboardEventHandler<HTMLDivElement>
  /**
   * The value to display when the input is empty.
   */
  placeholder?: string
  /**
   * The CSS styles to apply to the contenteditable div container.
   */
  style?: React.CSSProperties
  /**
   * The CSS class name to apply to the contenteditable div container.
   */
  className?: string
  /**
   * If provided, this function will be called to generate a completion instead of fetching one from the model.
   */
  mockCompletion?: (text: string) => string
}

/**
 * A text input that provides intelligent, tab-completable completions based on the text that has already been entered.
 * Due to the limitations of `<input>` and `<textarea>` elements, this component renders a `<div contenteditable="true">` instead.
 */
export function SmartInput(props: SmartInputProps) {
  const editor = React.useMemo(() => withHistory(withReact(createEditor())), [])
  const [completion, setCompletion] = React.useState<ICompletionResponse | null>(null)

  const [fetcher] = React.useState(
    new SmartInputFetcher(
      (text, completion) => {
        addCompletion(completion)
      },
      props.modelId,
      props.modelId ? props.mockCompletion : () => ' -- sample completion --'
    )
  )

  const acceptCompletion = () => {
    EditorHelper.acceptCompletion(editor, fetcher, completion)
    setCompletion(null)
  }

  const addCompletion = (completion: ICompletionResponse) => {
    if (props.value) {
      // don't add completion if the text is gone
      EditorHelper.addCompletion(editor, fetcher, completion)
      setCompletion(completion)
    }
  }

  const clearCompletion = () => {
    EditorHelper.removeCompletion(editor)
    setCompletion(null)
  }

  // for good measure, always overwrite the fetcher fields
  // fetcher.modelId = props.modelId
  fetcher.addCompletion = (text, completion) => {
    const currText = props.value
    if (completion && (text + completion).startsWith(currText)) {
      addCompletion(completion)
    }
  }
  // fetcher.mockCompletion = props.mockCompletion

  const initialValue: CustomDescendant[] = [
    {
      type: 'text',
      children: [{ text: props.value, type: undefined }],
    },
  ]

  useEffect(() => {
    if (props.editorRef) {
      props.editorRef.current = {
        ...ReactEditor.toDOMNode(editor, editor),
        focus: () => {
          setTimeout(() => {
            ReactEditor.focus(editor)
            if (props.value) {
              Transforms.select(editor, Editor.end(editor, []))
            } else {
              Transforms.select(editor, Editor.start(editor, []))
            }
          }, 50)
        },
      }
    }
  })

  const renderElement = React.useCallback(
    (p: any) => {
      switch (p.element.type) {
        case 'text':
          return props.renderText({
            ...p,
          })
        case 'completion':
          return props.renderCompletion({
            ...p,
            contentEditable: 'false',
            suppressContentEditableWarning: true,
          })
        default:
          return <DefaultElement {...p} />
      }
    },
    // eslint-disable-next-line react-hooks/exhaustive-deps
    [props.renderText, props.renderCompletion]
  )

  const renderPlaceholder = React.useCallback(
    (p: any) => {
      return props.renderPlaceholder!({
        ...p,
        ...(p.attributes || {}), // for some reason, Slate needs the attributes splatted differently on the placeholder for it to work correctly lmao
        style: {
          // these style attributes usually come from slate, we gotta make sure to add them back or the cursor might go after the placeholder lol
          position: 'absolute',
          pointerEvents: 'none',
          width: '100%',
          maxWidth: '100%',
          userSelect: 'none',
          textDecoration: 'none',
          ...(p.style || {}),
        },
      })
    },
    [props.renderPlaceholder]
  )

  return (
    <Slate
      editor={editor}
      value={initialValue}
      onChange={value => {
        const text = EditorHelper.serialize(value.filter(n => n.type !== 'completion'))
        fetcher.fetchCompletion(text, props.context || {})?.catch(e => console.error(e))
        fetcher.sendEvents(text, props.context || {})
        props.onChange(text)
      }}
    >
      <Editable
        placeholder={props.placeholder}
        className={props.className}
        style={{ ...(props.style || {}), whiteSpace: props.multiline ? undefined : 'pre' }}
        onPaste={(event: any) => {
          clearCompletion()
          event.preventDefault()
          const value = event.clipboardData.getData('Text')
          if (!props.multiline) {
            editor.insertText(value.replaceAll('\n', ''))
          } else {
            editor.insertText(value)
          }
        }}
        renderPlaceholder={props.renderPlaceholder ? renderPlaceholder : undefined}
        renderElement={renderElement}
        onBlur={e => {
          clearCompletion()
          fetcher.currentCompletion = null
          if (props.onBlur) {
            props.onBlur(e)
          }
        }}
        onFocus={e => {
          if (props.onFocus) {
            props.onFocus(e)
            if (editor.selection == null) {
              // fix bug for chrome
              if (props.value) {
                Transforms.select(editor, Editor.end(editor, []))
              } else {
                Transforms.select(editor, Editor.start(editor, []))
              }
            }
          }
        }}
        onKeyDown={(event: any) => {
          const hasCompletion = Boolean(completion?.completion?.trim()?.length)
          if (event.key === 'Tab' && hasCompletion) {
            event.preventDefault()
            acceptCompletion()
            fetcher.currentCompletion = null
          } else if (event.key === 'ArrowRight' && hasCompletion) {
            event.preventDefault()
            acceptCompletion()
            fetcher.currentCompletion = null
          } else if (event.key === 'ArrowLeft' && hasCompletion) {
            clearCompletion()
            fetcher.currentCompletion = null
          } else if ((event.key === 'ArrowDown' || event.key === 'ArrowUp') && hasCompletion) {
            clearCompletion()
          } else if (event.key === 'Backspace') {
            fetcher.currentCompletion = null
            clearCompletion()
          } else if (event.key === 'Enter' && !event.metaKey) {
            event.preventDefault()
            if (props.multiline) {
              clearCompletion()
              editor.insertText('\n')
            }
          } else if (event.key !== 'Shift') {
            if (
              completion &&
              !event.metaKey &&
              completion.completion?.startsWith(event.key) &&
              event.key.length === 1
            ) {
              const newCompletionText = completion.completion.substring(1)
              // don't clear the completion if the user is just typing the completion character for character
              EditorHelper.updateCompletion(editor, newCompletionText)
              setCompletion({ ...completion, completion: newCompletionText })
            } else {
              fetcher.currentCompletion = null
              clearCompletion() // TODO: this is a bit aggressive?
            }
          }

          if (props.onKeyDown) {
            props.onKeyDown(event)
          }
        }}
      />
    </Slate>
  )
}

const EditorHelper = {
  isCompletionActive: (editor: CEditor) => {
    const completion = editor.children.find(n => n.type === 'completion')
    return !!completion
  },

  isCursorAtEndOfDocument: (editor: CEditor) => {
    const sel = editor.selection
    if (!sel) {
      return false
    }
    const [node] = sel.anchor.path
    const offset = sel.anchor.offset
    if (node !== editor.children.length - 1) {
      // not at the last node
      return false
    }
    return offset === Node.string(editor.children[node]).length
  },

  acceptCompletion: (editor: CEditor, fetcher: SmartInputFetcher, completion: ICompletionResponse | null) => {
    const isActive = EditorHelper.isCompletionActive(editor)
    if (isActive && completion?.completion) {
      EditorHelper.removeCompletion(editor)
      editor.insertText(completion.completion)
      fetcher.updateCompletion(completion.sid, { accepted: true })
    }
  },

  addCompletion(editor: CEditor, fetcher: SmartInputFetcher, completion: ICompletionResponse) {
    const isActive = EditorHelper.isCompletionActive(editor)
    const isCursorEnd = EditorHelper.isCursorAtEndOfDocument(editor)
    if (!isActive && isCursorEnd && completion?.completion) {
      Transforms.insertNodes(
        editor,
        {
          type: 'completion',
          children: [{ text: completion.completion }],
        } as CustomDescendant,
        {
          at: [editor.children.length],
        }
      )
      fetcher.updateCompletion(completion.sid, { seen: true })
    }
  },

  removeCompletion(editor: CEditor) {
    const isActive = EditorHelper.isCompletionActive(editor)
    if (isActive) {
      Transforms.removeNodes(editor, {
        at: [editor.children.length - 1],
      })
    }
  },

  updateCompletion(editor: CEditor, newCompletion: string) {
    const isActive = EditorHelper.isCompletionActive(editor)
    if (isActive && newCompletion.length > 0) {
      // could not find a way to update to set the text of the node, so we remove the first character only
      Transforms.delete(editor, {
        at: {
          anchor: { path: [editor.children.length - 1, 0], offset: 0 },
          focus: { path: [editor.children.length - 1, 0], offset: 1 },
        },
      })
    }
  },

  serialize(nodes: CustomDescendant[]) {
    return nodes.map(n => Node.string(n)).join('\n')
  },
}

type CustomElement = {
  type: 'text' | 'completion' | undefined
  children: CustomText[]
}

type CustomText = { text: string; type: undefined }

type CustomDescendant = CustomElement | CustomText

type CEditor = Omit<BaseEditor, 'children'> & ReactEditor & { children: CustomDescendant[] }

declare module 'slate' {
  interface CustomTypes {
    Editor: CEditor
    Element: CustomElement
    Text: CustomText
  }
}
