import debounce from 'lodash.debounce'
import throttle from 'lodash.throttle'
import { ulid } from 'ulid'
import { ICompletionResponse } from './ICompletionResponse'

/**
 * Responsible for fetching completions from a model and adding them to the input.
 */
export class SmartInputFetcher {
  /**
   * sequenceId is used to ensure that we don't add completions from a previous fetch
   */
  sequenceId = 0

  sessionId = ulid()

  fetchCache: Record<string, ICompletionResponse | null> = {}

  alreadySeen = new Set<string>()
  alreadyAccepted = new Set<string>()

  /**
   * currentCompletion is used to track the latest text that was sent to the model for fetching completions, and the
   * completion that was returned (if any ~ we may be waiting for it).
   */
  currentCompletion: {
    sequenceId: number
    text: string
    completion?: string | null
  } | null = null

  constructor(
    public addCompletion: (text: string, completion: ICompletionResponse) => void,
    public modelId?: string,
    public mockCompletion?: (text: string) => string | null
  ) {}

  fetchCompletion = debounce(async (text: string, context: any) => {
    if (text.length === 0) {
      // only get completions for non-empty text
      return null
    }

    const seqId = ++this.sequenceId

    // if a completion was already added, and the user is typing the same characters as the current completion, don't fetch
    if (this.currentCompletion?.completion) {
      const currExpectedText = this.currentCompletion.text + this.currentCompletion.completion
      if (currExpectedText.startsWith(text)) {
        return null
      }
    }

    // update the current completion
    this.currentCompletion = { text, sequenceId: seqId }

    // if we already have a completion for this text, don't fetch, but do return it
    const cachedVal = this.fetchCache[text]
    if (cachedVal) {
      // don't fetch the same text twice
      this.addCompletion(text, cachedVal)
      this.currentCompletion.completion = cachedVal.completion
      return cachedVal
    }

    if (this.mockCompletion) {
      const completion = this.mockCompletion(text)
      if (completion) {
        return await new Promise<ICompletionResponse>(resolve => {
          setTimeout(() => {
            const c = { input: text, sid: ulid(), completion }
            this.fetchCache[text] = c
            if (this.currentCompletion?.sequenceId === seqId) {
              this.addCompletion(text, c)
              this.currentCompletion.completion = completion
            }
            resolve(c)
          }, getRandomInt(200, 1000)) // simulate API latency
        })
      }
      return { input: text, sid: ulid(), completion: null }
    }

    const res = await fetch('https://api.foundation-ui.com/completions', {
      method: 'POST',
      mode: 'cors',
      headers: {
        'Content-Type': 'application/json',
        Authorization: `Bearer ${this.modelId}`,
      },
      body: JSON.stringify({
        model: this.modelId,
        input: text,
        session: this.sessionId,
        context,
      }),
    })
    if (res.status !== 200) {
      console.error("Couldn't fetch completions")
      return null
    }
    const data = (await res.json()) as ICompletionResponse
    this.fetchCache[text] = data
    if (this.currentCompletion?.sequenceId === seqId) {
      this.addCompletion(text, data)
      this.currentCompletion.completion = data.completion
    }
    return data
  }, 200)

  updateCompletion(completionSid: string, arg: { seen?: boolean; accepted?: boolean }): void {
    if (this.mockCompletion) {
      return
    }

    // reduce chattiness
    const shouldUpdate =
      (arg.seen && !this.alreadySeen.has(completionSid)) || (arg.accepted && !this.alreadyAccepted.has(completionSid))
    if (!shouldUpdate) {
      return
    }

    fetch('https://api.foundation-ui.com/completions', {
      method: 'PUT',
      mode: 'cors',
      headers: {
        'Content-Type': 'application/json',
        Authorization: `Bearer ${this.modelId}`,
      },
      body: JSON.stringify({
        completionSid,
        seen: arg.seen,
        accepted: arg.accepted,
      }),
    })
      .then(res => {
        if (res.status !== 200) {
          console.error("Couldn't update completion")
        }
        if (arg.seen) {
          // only update the set if the update was successful
          this.alreadySeen.add(completionSid)
        }
        if (arg.accepted) {
          // only update the set if the update was successful
          this.alreadyAccepted.add(completionSid)
        }
      })
      .catch(e => console.error(e))
  }

  sendEvents = throttle(
    (text: string, context: any): void => {
      fetch('https://api.foundation-ui.com/events', {
        method: 'POST',
        mode: 'cors',
        headers: {
          'Content-Type': 'application/json',
          Authorization: `Bearer ${this.modelId}`,
        },
        body: JSON.stringify({
          model: this.modelId,
          session: this.sessionId,
          input: text,
          context,
        }),
      })
        .then(res => {
          if (res.status !== 200) {
            console.error('Error sending events')
          }
        })
        .catch(e => console.error(e))
    },
    1000,
    { trailing: true }
  )
}

/**
 * Returns a random integer between min (inclusive) and max (inclusive).
 * The value is no lower than min (or the next integer greater than min
 * if min isn't an integer) and no greater than max (or the next integer
 * lower than max if max isn't an integer).
 * Using Math.round() will give you a non-uniform distribution!
 */
function getRandomInt(min: number, max: number): number {
  min = Math.ceil(min)
  max = Math.floor(max)
  return Math.floor(Math.random() * (max - min + 1)) + min
}
